#ifndef Formule
#define Formule
#include <iostream>
// string est la classe des chaînes de caractères en C++ que l'on privilégie aux char * du C.
#include <string>
// set est la classe des ensembles de la STL du C++. set garantit l'unicité et
// le tri lors du parcours : le nombre et l'ordre des insertions ne compte pas.
// L'accès direct avec [] n'est pas possible : l'utilisation des itérateurs est obligatoire.
#include <set>
// vector est la classe des tableaux dynamiques de la STL du C++. vector ne 
// garantit PAS l'unicité et le tri lors du parcours : le nombre et l'ordre
// des insertions compte. L'accès direct avec [] est possible.
#include <vector>
// map est la classe des tableaux associatifs de la STL du C++.
#include <map>

using namespace std;

// Identificateurs pour les connecteurs de la logique propositionnelle : 
// ce sont les symboles associés aux nœuds de l'arbre de syntaxe des formules
enum operateur {
  o_variable,
  o_non,
  o_et,
  o_ou,
  o_implique,
  o_equivaut
};

// Structure des nœuds de l'arbre de syntaxe abstraite des formules
struct formule_s {
  // A chaque noeud est associé un opérateur
  operateur op;
  union {
    // Pour le cas o_variable, on associe une chaîne pour identifier la proposition
    string * nom;
    // Pour le cas du connecteur unaire o_non, on associe un seul sous-arbre arg.
    formule_s * arg;
    // Pour les cas des connecteurs binaires (o_et, o_ou, o_implique et o_equivaut),
    // on associe deux sous-arbres : le fils gauche arg1 et le fils droit arg2.
    struct {
      formule_s * arg1;
      formule_s * arg2;
    };
  };
};

typedef struct formule_s * formule;

// Création de formules simples ////////////////////////////////////////////////
formule variable(string nom);
formule variable(char* nom);
formule non(formule arg);
formule et(formule arg1, formule arg2);
formule ou(formule arg1, formule arg2);
formule impl(formule arg1, formule arg2);
formule equiv(formule arg1, formule arg2);
// Affichage d'une formule par induction sur son arbre de syntaxe //////////////
string  formule2string(formule f);
ostream& operator<<(ostream& out, formule f);
// Gestion de la mémoire dynamique des formules ////////////////////////////////
formule copier(formule f);
void free_all(formule f);

// Types des substitutions /////////////////////////////////////////////////////
typedef map<string,formule> subs_t;

// Types liés aux CNFs utilisées pour la résolution ////////////////////////////
// Type des variables
typedef int  var_t;
// Type des littéraux 
typedef unsigned int lit_t;
// Type des clauses, définies comme des ensembles de littéraux
typedef set<lit_t>    cls_t;
// Type des formules normales conjonctives (CNF), définies comme des tableaux
// dynamiques (vector en C++) de clauses
typedef vector<cls_t> cnf_t;

// fonction utilisée dans le main pour interagir avec minisat //////////////////
void writeDimacs(ostream&out, const map<string,int> & corresp, const cnf_t& cnf);

////////////////////////////////////////////////////////////////////////////////
//        A PARTIR D'ICI LES FONCTIONS SONT SIMPLEMENT DÉCLARÉES ET           //
//               DOIVENT ÊTRE IMPLÉMENTÉES DANS FORMULE.CPP                   //
////////////////////////////////////////////////////////////////////////////////

// Fonctions de manipulation des CNF  //////////////////////////////////////////

// Affichage d'une clause
ostream& operator<<(ostream& out, const cls_t& cl);

// Affichage d'une CNF
ostream& operator<<(ostream& out, const cnf_t& c);

// Renvoie vrai si la clause est dans la CNF
bool est_dans(const cls_t& cl, const cnf_t& c);

// Renvoie un littéral correspondant à une variable
// Le littéral sera positif si p vaut true, négatif sinon
lit_t var2lit(var_t v, bool p = true);

// Renvoie la variable correspondant à un littéral
var_t lit2var(lit_t l);

// Renvoie vrai si le littéral est positif
bool positif(lit_t l);

// Renvoie le littéral opposé 
lit_t neg(lit_t l);


// Fonctions de manipulation des formules //////////////////////////////////////

// Applique une substitution sur une formule.
// Il est demandé de copier systématiquement les formules lors de la
// construction du résultat afin de faciliter la gestion mémoire.
formule applique(formule f, const subs_t & sub);

// Numérote les variables apparaissant dans une formule afin de
// préparer la transformation en CNF.
void numerote(formule f, map<string,var_t> & corresp);


// Fonctions de transformation en CNF //////////////////////////////////////////
lit_t tseitin(formule f, map<string,var_t> & corresp, cnf_t& cnf, var_t & next); 

void tseitin(formule f, map<string,var_t> & corresp, cnf_t& cnf);

#endif
